#!/bin/bash
# vim:et ts=4 sts=4:
#
# LICENSE {{{
# The BSD License (http://www.opensource.org/licenses/bsd-license.php)
# specifies the terms and conditions of use for checksec.sh:
#
# Copyright 2013, pi-rho <ubuntu@tyr.cx>
# Copyright 2009-2011, Tobias Klein.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
# * Neither the name of Tobias Klein nor the name of trapkit.de may be
#   used to endorse or promote products derived from this software
#   without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#
# }}}
#-- global vars {{{
have_readelf=1
verbose=false
VERSION="1.5+"
PKGDATE="August 2013"

# FORTIFY_SOURCE vars
FS_end=_chk
FS_cnt_total=0
FS_cnt_checked=0
FS_cnt_unchecked=0
FS_chk_func_libc=0
FS_functions=0
FS_libc=0

# awk scripts
AWK_clean_chk_funcs='
/_chk@@/ {
    sub(/__/, ""); sub(/_chk@.*/, "");
    z[$8]++
}
END { for (i in z) print i | "sort" }'
AWK_clean_funcs='
/FUNC/ {
    sub(/__/, ""); sub(/@.*/, "");
    z[$8]++
}
END { for (i in z) print i | "sort" }'

# colors
FAIL='\033[31m'
REVR='\033[37;41m'
GOOD='\033[32m'
WARN='\033[33m'
RST='\033[0m'

#-- global vars }}}
#-- output functions {{{

# printerr [errno] "message"
printerr() {
    if isNumeric $1; then err=$1; shift; fi
    printf "${FAIL}Error: $*${RST}\n"
    [[ -z ${err} ]] || exit ${err}
}

# printxxx [width] "message" "ending"
# -- print a message with color
#    printxxx 20 "Good things" "\n\n"
#    printxxx "Good things" "\n\n"
printgood() {
    if isNumeric ${1}; then local width=-${1}; shift; fi
    printf "${GOOD}%${width}s${RST}${2}" "${1}"
}

printwarn() {
    if isNumeric ${1}; then local width=-${1}; shift; fi
    printf "${WARN}%${width}s${RST}${2}" "${1}"
}

printfail() {
    if isNumeric ${1}; then local width=-${1}; shift; fi
    printf "${FAIL}%${width}s${RST}${2}" "${1}"
}

#-- output functions }}}
#-- require routines {{{

requires_readelf() {
    [[ ${have_readelf} -ne 0 ]] || exit 1
}

requires_procread() {
    local pid=${1}
    if [[ ! -r /proc/${pid}/exe ]]; then
        root_privs || printerr 1 "No read permissions for '/proc/${pid}/exe' (run as root)."

        if [[ ! $(readlink /proc/${pid}/exe) ]]; then
            printerr 1 "Permission denied. Requested process ID belongs to a kernel thread."
        fi
        exit 1
    fi
}

requires_fileread() {
    local file=${1}
    [[ -n ${file} ]] || printerr 1 "Please provide a valid filename."
    [[ -e ${file} ]] || printerr 1 "The file '${file}' does not exist."
    [[ -r ${file} ]] || printerr 1 "No read permissions for '${file}' (run as root)."
}

requires_fileelf() {
    local file=${1}
    out=$(file ${file})
    [[ $out =~ ELF ]] || printerr 1 "'${file}' is not an ELF file: ${out}"
}

requires_pidvalid() {
    local pid=${1}
    [[ -n ${pid} ]] || printerr 1 "Please provide a valid process ID."
    isNumeric ${pid} || printerr 1 "Please provide a valid process ID."
    local pids=( $(cd /proc; ls -d [0-9]*) )
    [[ ${pids[@]} =~ \ ${pid}\  ]] || printerr 1 "Please provide a valid process ID."
}

requires_procvalid() {
    [[ -n ${optarg} ]] || printerr 1 "Please provide a valid process name."
    isString ${optarg} || printerr 1 "Please provide a valid process name."
}

#-- require routines }}}
#-- utility functions {{{

# version information
version() {
    printf "${GOOD}checksec ${VERSION}${RST}, ${PKGDATE}\n\n"
}

# help
help() {
    cat <<EOB
Usage: checksec [OPTION]

Options:

  --file <executable-file>
  --dir <directory> [-v]
  --file-libs <executable-file>
  --proc <process name>
  --proc-all
  --proc-libs <process ID>
  --kernel
  --fortify-file <executable-file>
  --fortify-proc <process ID>
  --version
  --help

EOB
}

# check if command exists
command_exists () {
    type $1 > /dev/null 2>&1
}

# check if directory exists
dir_exists () {
    [[ -d $1 ]]
}

# check user privileges
root_privs () {
    [[ ${UID} -eq 0 ]]
}

# check if input is numeric
isNumeric () {
    [[ ${1} =~ ^[0-9]+$ ]]
}

# check if input is a string
isString () {
    [[ ${1} =~ ^[A-Za-z]+$ ]]
}

#-- utility functions }}}
#-- readelf-based checks {{{

#---- file {{{

print_file_head() {
    printf "RELRO           STACK CANARY      NX            PIE             RPATH      RUNPATH      FILE\n"
}

filecheck() {
    local file=${1}
    header=$(readelf -h ${file} 2>/dev/null)
    dynamic=$(readelf -d ${file} 2>/dev/null)

    w=16  # COLUMN: RELRO
    if [[ $(readelf -l ${file} 2>/dev/null) =~ GNU_RELRO ]]; then
        if [[ ${dynamic} =~ BIND_NOW ]]; then
            printgood $w "Full RELRO"
        else
            printwarn $w "Partial RELRO"
        fi
    else
        printfail $w "No RELRO"
    fi

    w=18  # COLUMN: STACK CANARY
    if [[ $(readelf -s ${file} 2>/dev/null) =~ __stack_chk_fail ]]; then
        printgood $w "Canary found"
    else
        printfail $w "No canary found"
    fi

    w=14  # COLUMN: NX
    if [[ $(readelf -W -l ${file} 2>/dev/null) =~ GNU_STACK\ .*\ RWE ]]; then
        printfail $w "NX disabled"
    else
        printgood $w "NX enabled"
    fi

    w=16  # COLUMN: PIE
    if [[ ${header} =~ Type:\ *EXEC ]]; then
        printfail $w "No PIE"
    elif [[ ${header} =~ Type:\ *DYN ]]; then
        if [[ ${dynamic} =~ \(DEBUG\) ]]; then
            printgood $w "PIE enabled"
        else
            printwarn $w "DSO"
        fi
    else
        printwarn $w "Not an ELF file"
    fi

    w=11  # COLUMN: RPATH
    if [[ ${dynamic} =~ rpath ]]; then
        printfail $w "RPATH"
    else
        printgood $w "No RPATH"
    fi

    w=13  # COLUMN: RUNPATH
    if [[ ${dynamic} =~ runpath ]]; then
        printfail $w "RUNPATH"
    else
        printgood $w "No RUNPATH"
    fi

    # COLUMN: File
    print_file_slug ${file}
}

print_file_slug() {
    local file=${1}
    if [[ -u ${file} || -g ${file} ]]; then
        printf "${REVR}${file}${RST}\n"
    else
        printf "${file}\n"
    fi
}

filelibcheck() {
    local file=${1}
    libs=( $(ldd --verbose ${file} | perl -ne 'chomp;m!(/[^ :]+)! and print $1,$/;' | sort -u) )

    printf "\nLinked libraries:\n\n"
    print_file_head
    for element in ${libs[@]}; do
        [[ ${element} != ${file} ]] || continue
        [[ -f ${element} ]] || continue
        filecheck ${element}
    done
}

#---- file }}}
#---- process {{{

print_proc_head() {
    printf "* Process information:\n\n"
    printf "         COMMAND    PID RELRO             STACK CANARY           NX/PaX        PIE\n"
}

print_proc_stub() {
    local pid=${1}
    printf "%16s%7d " $(awk '/Name:/ {print $2}' /proc/${pid}/status) ${pid}
}

proccheck() {
    pheaders=$(readelf -W -l $1/exe 2>/dev/null)
    dynamic=$(readelf -d $1/exe 2>/dev/null)

    w=18  # COLUMN: RELRO
    if [[ ${pheaders} =~ Program\ Headers ]]; then
        if [[ ${pheaders} =~ GNU_RELRO ]]; then
            if [[ ${dynamic} =~ BIND_NOW ]]; then
                printgood $w "Full RELRO"
            else
                printwarn $w "Partial RELRO"
            fi
        else
            printfail $w "No RELRO"
        fi
    else
        printerr 1 "Permission denied (please run as root)"
    fi

    symbols=$(readelf -s $1/exe 2>/dev/null)

    w=23  # COLUMN: STACK CANARY
    if [[ ${symbols} =~ Symbol\ table ]]; then
        if [[ ${symbols} =~ __stack_chk_fail ]]; then
            printgood $w "Canary found"
        else
            printfail $w "No canary found"
        fi
    else
        if [[ "$1" != "1" ]]; then
            printwarn $w "Permission denied"
        else
            printwarn $w "No symbol table found"
        fi
    fi

    w=14  # COLUMN: NX/PaX
    status=$(awk '/^PaX:/ {print $2}' $1/status 2>/dev/null)
    if [[ -n ${status} ]]; then
        case ${status} in
            P?MRs|p?MRS|P?MRS) printgood $w "PaX enabled"   ;;
            p??Rs)             printwarn $w "PaX ASLR only" ;;
            P?mRs|p?mRS|P?mRS) printwarn $w "PaX mprot off" ;;
            P?Mrs|p?MrS|P?MrS) printwarn $w "PaX ASLR off"  ;;
            P?mrs|p?mrS|P?mrS) printwarn $w "PaX NX only"   ;;
            *)                 printfail $w "PaX disabled"  ;;
        esac
    elif [[ ${pheaders} =~ GNU_STACK\ .*\ RWE ]]; then
        printfail $w "NX disabled"
    else
        printgood $w "NX enabled"
    fi

    headers=$(readelf -h $1/exe 2>/dev/null)
    w=24  # COLUMN: PIE
    if [[ ${headers} =~ Type:\ *EXEC ]]; then
        printfail $w "No PIE"
    elif [[ ${headers} =~ Type:\ *DYN ]]; then
        if [[ ${dynamic} =~ \(DEBUG\) ]]; then
            printgood $w "PIE enabled"
        else
            printwarn $w "Dynamic Shared Object"
        fi
    else
        printwarn $w "Not an ELF file"
    fi
    echo
}

proclibcheck() {
    local pid=${1}
    # $5 = inode, $6 = pathname
    libs=( $(awk '$5>0 {z[$6]++}; END {for (i in z) print i}' /proc/${pid}/maps | xargs file | awk -F: '/ELF/ { print $1 }') )

    printf "\n* File and loaded libraries (# of mapped files: ${#libs[@]}):\n\n  "
    print_file_head
    for element in ${libs[@]}; do
        printf "  "
        filecheck ${element}
    done
}

#---- process }}}

#-- readelf-based checks }}}
#-- proc and sysctl-based checks {{{

aslrcheck() {
    printf "* System-wide ASLR"
    # PaX ASLR support
    status=$(</proc/1/status)
    if [[ ! ${status} =~ Name: ]]; then
        printwarn " insufficient privileges for PaX ASLR checks" "\n"
        printf "  Fallback to standard Linux ASLR check"
    fi

    if [[ ${status} =~ PaX: ]]; then
        printf ": "
        if [[ ${status} =~ PaX:.*R ]]; then
            printgood "PaX ASLR enabled" "\n\n"
        else
            printfail "PaX ASLR disabled" "\n\n"
        fi
    else
        # standard Linux 'kernel.randomize_va_space' ASLR support
        # (see the kernel file 'Documentation/sysctl/kernel.txt' for a detailed description)
        printf " (kernel.randomize_va_space): "
        krvs=$(/sbin/sysctl -a 2>/dev/null | awk '/kernel\.randomize_va_space =/ {print $3}')
        if [[ ${krvs} -eq 1 ]]; then
            printwarn "On (Setting: 1)" "\n\n"
            printf "  Description - Make the addresses of mmap base, stack and VDSO page randomized.\n"
            printf "  This, among other things, implies that shared libraries will be loaded to \n"
            printf "  random addresses. Also for PIE-linked binaries, the location of code start\n"
            printf "  is randomized. Heap addresses are *not* randomized.\n\n"
        elif [[ ${krvs} -eq 2 ]]; then
            printgood "On (Setting: 2)" "\n\n"
            printf "  Description - Make the addresses of mmap base, heap, stack and VDSO page randomized.\n"
            printf "  This, among other things, implies that shared libraries will be loaded to random \n"
            printf "  addresses. Also for PIE-linked binaries, the location of code start is randomized.\n\n"
        elif [[ ${krvs} -eq 0 ]]; then
            printfail "Off (Setting: 0)" "\n"
        else
            printfail "Not supported" "\n"
        fi
        printf "  See the kernel file 'Documentation/sysctl/kernel.txt' for more details.\n\n"
    fi
}

nxcheck() {
    printf "* Does the CPU support NX: "
    if [[ $(</proc/cpuinfo) =~ nx ]]; then
        printgood "Yes" "\n\n"
    else
        printfail "No" "\n\n"
    fi
}

kernelcheck() {
    printf "* Kernel protection information:\n\n"
    printf "  Description - List the status of kernel protection mechanisms. Rather than\n"
    printf "  inspect kernel mechanisms that may aid in the prevention of exploitation of\n"
    printf "  userspace processes, this option lists the status of kernel configuration\n"
    printf "  options that harden the kernel itself against attack.\n\n"
    printf "  Kernel config: "

    cd /proc
    if [[ -f /proc/config.gz ]]; then
        kconfig=$(zcat /proc/config.gz)
        printgood "/proc/config.gz" "\n\n"
    elif [[ -f /boot/config-$(uname -r) ]]; then
        kconfig=$(</boot/config-$(uname -r))
        printf "${WARN}/boot/config-$(uname -r)${RST}\n\n"
        printf "  Warning: The config on disk may not represent running kernel config!\n\n";
    elif [[ -f "${KBUILD_OUTPUT:-/usr/src/linux}"/.config ]]; then
        kconfig=$(<${KBUILD_OUTPUT:-/usr/src/linux}/.config)
        printf "${WARN}%s${RST}\n\n" "${KBUILD_OUTPUT:-/usr/src/linux}/.config"
        printf "  Warning: The config on disk may not represent running kernel config!\n\n";
    else
        printfail "NOT FOUND" "\n\n"
        exit 0
    fi

    printf "  GCC stack protector support:            "
    if [[ ${kconfig} =~ CONFIG_CC_STACKPROTECTOR=[Yy] ]]; then
        printgood "Enabled" "\n"
    else
        printfail "Disabled" "\n"
    fi

    printf "  Strict user copy checks:                "
    if [[ ${kconfig} =~ CONFIG_DEBUG_STRICT_USER_COPY_CHECKS=[Yy] ]]; then
        printgood "Enabled" "\n"
    else
        printfail "Disabled" "\n"
    fi

    printf "  Enforce read-only kernel data:          "
    if [[ ${kconfig} =~ CONFIG_DEBUG_RODATA=[Yy] ]]; then
        printgood "Enabled" "\n"
    else
        printfail "Disabled" "\n"
    fi
    printf "  Restrict /dev/mem access:               "
    if [[ ${kconfig} =~ CONFIG_STRICT_DEVMEM=[Yy] ]]; then
        printgood "Enabled" "\n"
    else
        printfail "Disabled" "\n"
    fi

    printf "  Restrict /dev/kmem access:              "
    if [[ ${kconfig} =~ CONFIG_DEVKMEM=[Yy] ]]; then
        printfail "Disabled" "\n"
    else
        printgood "Enabled" "\n"
    fi

    printf "\n"
    printf "* grsecurity / PaX: "

    if [[ ${kconfig} =~ CONFIG_GRKERNSEC=[Yy] ]]; then
        if [[ ${kconfig} =~ CONFIG_GRKERNSEC_HIGH=[Yy] ]]; then
            printgood "High GRKERNSEC" "\n\n"
        elif [[ ${kconfig} =~ CONFIG_GRKERNSEC_MEDIUM=[Yy] ]]; then
            printwarn "Medium GRKERNSEC" "\n\n"
        elif [[ ${kconfig} =~ CONFIG_GRKERNSEC_LOW=[Yy] ]]; then
            printfail "Low GRKERNSEC" "\n\n"
        else
            printwarn "Custom GRKERNSEC" "\n\n"
        fi

        printf "  Non-executable kernel pages:            "
        if [[ ${kconfig} =~ CONFIG_PAX_KERNEXEC=[Yy] ]]; then
            printgood "Enabled" "\n"
        else
            printfail "Disabled" "\n"
        fi

        printf "  Prevent userspace pointer deref:        "
        if [[ ${kconfig} =~ CONFIG_PAX_MEMORY_UDEREF=[Yy] ]]; then
            printgood "Enabled" "\n"
        else
            printfail "Disabled" "\n"
        fi

        printf "  Prevent kobject refcount overflow:      "
        if [[ ${kconfig} =~ CONFIG_PAX_REFCOUNT=[Yy] ]]; then
            printgood "Enabled" "\n"
        else
            printfail "Disabled" "\n"
        fi

        printf "  Bounds check heap object copies:        "
        if [[ ${kconfig} =~ CONFIG_PAX_USERCOPY=[Yy] ]]; then
            printgood "Enabled" "\n"
        else
            printfail "Disabled" "\n"
        fi

        printf "  Disable writing to kmem/mem/port:       "
        if [[ ${kconfig} =~ CONFIG_GRKERNSEC_KMEM=[Yy] ]]; then
            printgood "Enabled" "\n"
        else
            printfail "Disabled" "\n"
        fi

        printf "  Disable privileged I/O:                 "
        if [[ ${kconfig} =~ CONFIG_GRKERNSEC_IO=[Yy] ]]; then
            printgood "Enabled" "\n"
        else
            printfail "Disabled" "\n"
        fi

        printf "  Harden module auto-loading:             "
        if [[ ${kconfig} =~ CONFIG_GRKERNSEC_MODHARDEN=[Yy] ]]; then
            printgood "Enabled" "\n"
        else
            printfail "Disabled" "\n"
        fi

        printf "  Hide kernel symbols:                    "
        if [[ ${kconfig} =~ CONFIG_GRKERNSEC_HIDESYM=[Yy] ]]; then
            printgood "Enabled" "\n"
        else
            printfail "Disabled" "\n"
        fi
    else
        printfail "No GRKERNSEC" "\n\n"
        printf "  The grsecurity / PaX patchset is available here:\n"
        printf "    http://grsecurity.net/\n"
    fi

    printf "\n"
    printf "* Kernel Heap Hardening: "

    if [[ ${kconfig} =~ CONFIG_KERNHEAP=[Yy] ]]; then
        if [[ ${kconfig} =~ CONFIG_KERNHEAP_FULLPOISON=[Yy] ]]; then
            printgood "Full KERNHEAP" "\n\n"
        else
            printwarn "Partial KERNHEAP" "\n\n"
        fi
    else
        printfail "No KERNHEAP" "\n\n"
        printf "  The KERNHEAP hardening patchset is available here:\n"
        printf "    https://www.subreption.com/kernheap/\n\n"
    fi
}

#-- proc and sysctl-based checks }}}
#-- FORTIFY_SOURCE related checks {{{

FS_libc_check() {
    printf "* FORTIFY_SOURCE support available (libc)    : "

    if [[ ${#FS_chk_func_libc[@]} -gt 0 ]]; then
        printgood "Yes" "\n"
    else
        printfail "No" "\n"
        exit 1
    fi
}

FS_binary_check() {
    printf "* Binary compiled with FORTIFY_SOURCE support: "

    for FS_tmp_func in ${FS_functions[@]}; do
        if [[ ${FS_tmp_func} =~ _chk$ ]]; then
            printgood "Yes" "\n"
            return
        fi
    done
    printfail "No" "\n"
}

FS_comparison() {
    echo
    printf " ------ EXECUTABLE-FILE ------- . -------- LIBC --------\n"
    printf " FORTIFY-able library functions | Checked function names\n"
    printf " -------------------------------------------------------\n"

    for FS_tmp_func in ${FS_functions[@]}; do
        # if the base function doesn't occur in the libc list, don't iterate
        [[ ${FS_chk_func_libc[@]} =~ ${FS_tmp_func/$FS_end} ]] || continue
        for FS_tmp_libc in ${FS_chk_func_libc[@]}; do
            if [[ $FS_tmp_func =~ ^$FS_tmp_libc$ ]]; then
                printf " ${FAIL}%-30s${RST} | __%s%s\n" $FS_tmp_func $FS_tmp_libc $FS_end
                let FS_cnt_total++
                let FS_cnt_unchecked++
            elif [[ $FS_tmp_func =~ ^$FS_tmp_libc($FS_end) ]]; then
                printf " ${GOOD}%-30s${RST} | __%s%s\n" $FS_tmp_func $FS_tmp_libc $FS_end
                let FS_cnt_total++
                let FS_cnt_checked++
            fi
        done
    done
}

FS_summary() {
    printf "\nSUMMARY:\n\n"
    printf "* Number of checked functions in libc                : %5d\n" ${#FS_chk_func_libc[@]}
    printf "* Total number of library functions in the executable: %5d\n" ${#FS_functions[@]}
    printf "* Number of FORTIFY-able functions in the executable : ${WARN}%5d${RST}\n" $FS_cnt_total
    printf "* Number of checked functions in the executable      : ${GOOD}%5d${RST}\n" $FS_cnt_checked
    printf "* Number of unchecked functions in the executable    : ${FAIL}%5d${RST}\n" $FS_cnt_unchecked
    echo
}

#-- FORTIFY_SOURCE related checks }}}
#-- main execution {{{

if ! command_exists readelf; then
    have_readelf=0
    printwarn "Warning: 'readelf' not found! It's required for most checks." "\n\n"
fi

# parse command-line arguments
case "$1" in
    --version) version ;;

    --help) help ;;

    --dir)  # --dir {{{
        optarg=${2}
        requires_readelf
        [[ ${3} == "-v" ]] && verbose=true
        [[ -n ${optarg} ]] || printerr 1 "Please provide a valid directory."
        tempdir=$(readlink -e ${optarg})  # canonicalize
        [[ -n $tempdir && -d $tempdir ]] || printerr 1 "The directory '$tmpdir' does not exist."

        cd $tempdir
        print_file_head
        for N in *; do
            [[ -f $N ]] || continue
            if $verbose && [[ -L $N ]]; then  # symbolic link
                link=$(readlink -e $N)
                printf "%47s ${FAIL}%-15s${RST} %-23s ${WARN}%s${RST}\n" "" "Error:" "symbolic link" "$N -> ${link//$tempdir\/}"
                continue
            fi
            if [[ -r $N ]]; then
                out=$(file $N)
                if [[ $out =~ ELF ]]; then
                    filecheck $N
                else
                    if $verbose; then
                        printf "%47s ${FAIL}%-15s${RST} %-23s ${FAIL}%s${RST}\n" "" "Error:" "$(file -b --mime-type $N)" "$N"
                    fi
                fi
            else
                $verbose && printerr "No read permissions for '$tempdir/$N' (run as root)."
            fi
        done
        ;;  # --dir }}}

    --file)  # --file {{{
        optarg=${2}
        requires_readelf
        requires_fileread ${optarg}
        requires_fileelf ${optarg}

        print_file_head
        filecheck ${optarg}
        ;;  # --file }}}

    --file-libs)  # --file-libs {{{
        optarg=${2}
        requires_readelf
        requires_fileread ${optarg}

        print_file_head
        filecheck ${optarg}
        filelibcheck ${optarg}
        ;;  # --file-libs }}}

    --proc-all)  # --proc-all {{{
        requires_readelf
        cd /proc
        aslrcheck
        nxcheck
        print_proc_head
        for N in [1-9]*; do
            [[ $N -ne $$ ]] || continue
            [[ -r $N/exe ]] || continue  # skip unreadable procs
            print_proc_stub $N
            proccheck $N
        done
        if ! root_privs; then
            printf "\n${WARN}Note: You are running 'checksec.sh' as an unprivileged user.\n"
            printf "      Too see all processes, please run the script as root.${RST}\n\n"
        fi
        ;;  # --proc-all }}}

    --proc)  # --proc {{{
        optarg=${2}
        requires_readelf
        requires_procvalid ${optarg}
        cd /proc
        aslrcheck
        nxcheck
        print_proc_head
        for N in $(pgrep ${optarg}); do
            print_proc_stub $N
            requires_procread $N
            proccheck $N
        done
        ;;  # --proc }}}

    --proc-libs)  # --proc-libs {{{
        optarg=${2}
        requires_readelf
        requires_pidvalid ${optarg}
        cd /proc
        aslrcheck
        nxcheck
        print_proc_head
        print_proc_stub ${optarg}
        requires_procread ${optarg}
        proccheck ${optarg}
        proclibcheck ${optarg}
        ;;  # --proc-libs }}}

    --kernel)  # --kernel {{{
        kernelcheck
        ;;  # --kernel }}}

    --fortify-file)  # --fortify-file {{{
        optarg=${2}
        requires_readelf
        requires_fileread ${optarg}
        requires_fileelf ${optarg}

        FS_libc=$(ldd ${optarg} | awk '/libc\.so\.[0-9]/ {print $3}')
        [[ -n ${FS_libc} && -r ${FS_libc} ]] || printerr 1 "LIBC not found."

        printf "* File name                                  : %s\n" ${optarg}
        FS_chk_func_libc=( $(readelf -s $FS_libc | awk "${AWK_clean_chk_funcs}") )
        FS_functions=( $(readelf -s ${optarg} | awk "${AWK_clean_funcs}") )

        FS_libc_check
        FS_binary_check
        FS_comparison
        FS_summary
        ;;  # --fortify-file }}}

    --fortify-proc)  # --fortify-proc {{{
        optarg=${2}
        requires_readelf
        requires_pidvalid ${optarg}

        cd /proc
        requires_procread ${optarg}

        FS_libc=$(ldd ${optarg}/exe | awk '/libc\.so\.6/ {print $3}')
        [[ -n ${FS_libc} && -r ${FS_libc} ]] || printerr 1 "LIBC not found."

        printf "* Process name (PID)                         : %s (%d)\n" $(awk '/Name:/ {print $2}' ${optarg}/status) ${optarg}
        FS_chk_func_libc=( $(readelf -s $FS_libc | awk "${AWK_clean_chk_funcs}") )
        FS_functions=( $(readelf -s ${optarg}/exe | awk "${AWK_clean_funcs}") )

        FS_libc_check
        FS_binary_check
        FS_comparison
        FS_summary
        ;;  # --fortify-proc }}}

    *)
        [[ "$#" = "0" ]] || printerr "Unknown option '$1'."
        help
        exit 1
        ;;
esac
exit 0
#-- main execution }}}
